.baz {
  color: red; }
  .baz {
    color: pink; }
  .baz--alert {
    color: orange; }
  .baz .bar {
    color: green; }
  .baz .foo {
    color: black; }
@keyframes kf2 {
  from: 0;
  to: 100; }
  @media print {
    .baz .bar {
      color: yellow; } }
  .baz {
    color: pink; }

  .baz--alert {
    color: orange; }

  .bar {
    color: blue; }

  .foo {
    color: white; }

@keyframes kf {
  from: 100;
  to: 0; }
  @media print {
    .bar {
      color: yellow; } }
