@mixin ie-opacity($opacity) {
  opacity: $opacity / 100;
  filter: alpha(opacity=$opacity);
  bilter: alpha(opacity=$opacity);
  kilter: type-of(opacity=$opacity);
  left: expression(document.body.clientWidth/2-oDiv.offsetWidth/2);
  flop: expression(document.body.clientHeight/2-oDiv.offsetHeight/2);
}

$startColor: red;
$endColor: green;

foo {
  filter: progid:Microsoft.foo.bar.Baz(flip=#{foo + bar}, bang=#00ff00cc);
  something: blah(hux = mumble);
  blah: progid:something.something(flip=foobar, bang=#abc);
  blah: progid:bar.hux();
  blah: type-of(hux = mumble);
  @include ie-opacity(.5);
  left: expression(document.body.clientWidth/4);
  filter: progid:DXImageTransform.Microsoft.gradient(startColorstr='#{ie-hex-str($startColor)}', endColorstr='#{ie-hex-str($endColor)}', GradientType=1);
}